/********************************************************************************
*                                                                               *
*                      F X   M O V I E   M A N A G E R                          *
*                                                                               *
*********************************************************************************
* Copyright(C) 2008-2011 by Luigi Ziviani <ziviani_luigi@yahoo.it>.             *
* All Rights Reserved.                                 v 5.3                    *
*********************************************************************************
* Id: fxplayer.c,                                      2008/01/01               *
*********************************************************************************
* This file is part of FXMOVIEMANAGER                                           *
* FXMOVIEMANAGER is free software; you can redistribute it and/or modify        *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation; either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* This program is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
* You should have received a copy of the GNU General Public License             *
* along with this program; if not, write to the Free Software                   *
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA     * 
********************************************************************************/


            /****************************************************
            * the code written in this file is based on ffplay, *
            * copyright (c) 2003 Fabrice Bellard,               *
            *                                                   *
            * I found also useful a tutorial by Stephen Dranger *
            * who updated the original tutorial by Martin Bohme *
            ****************************************************/


#include "config.h"

#ifdef HAVE_INTTYPES_H
 #include <inttypes.h>
#endif

#include <math.h>

#include <libavcodec/avcodec.h>
#include <libavformat/avformat.h>
#include <libswscale/swscale.h>

#include <SDL/SDL.h>
#include <SDL/SDL_thread.h>

#ifdef __MINGW32__
 #undef main /* Prevents SDL from overriding main() */
#endif


/* SDL audio buffer size, in samples. Should be small to have precise
   A/V sync as SDL does not have hardware buffer fullness info. */
#define SDL_AUDIO_BUFFER_SIZE 1024

#define MAX_VIDEOQ_SIZE (5 * 256 * 1024)
#define MAX_AUDIOQ_SIZE (5 * 16 * 1024)
/* no AV sync correction is done if below the AV sync threshold */
#define AV_SYNC_THRESHOLD 0.01
/* no AV correction is done if too big error */
#define AV_NOSYNC_THRESHOLD 10.0
/* maximum audio speed change to get correct sync */
#define SAMPLE_CORRECTION_PERCENT_MAX 10
/* we use about AUDIO_DIFF_AVG_NB A-V differences to make the average */
#define AUDIO_DIFF_AVG_NB 20

#define FF_ALLOC_EVENT   (SDL_USEREVENT)
#define FF_REFRESH_EVENT (SDL_USEREVENT + 1)
#define FF_QUIT_EVENT    (SDL_USEREVENT + 2)

#define VIDEO_PICTURE_QUEUE_SIZE 1

#define DEFAULT_AV_SYNC_TYPE AV_SYNC_VIDEO_MASTER

static int sws_flags = SWS_BICUBIC;

typedef struct PacketQueue {
  AVPacketList *first_pkt, *last_pkt;
  int quit;
  int nb_packets;
  int size;
  SDL_mutex *mutex;
  SDL_cond *cond;
} PacketQueue;

typedef struct VideoPicture {
  double pts;
  SDL_Overlay *bmp;
  int width, height; /* source height & width */
  int allocated;
  enum PixelFormat pix_fmt;
} VideoPicture;

typedef struct VideoState {

  AVFormatContext *avf_ctx;

  int             video_stream; 
  int             audio_stream;

  int             av_sync_type;
  double          external_clock; /* external clock base */
  int64_t         external_clock_time;

  int             seek_req;
  int             seek_flags;
  int64_t         seek_pos;
  int64_t         seek_rel;

  double          audio_clock;
  AVStream        *audio_st;
  PacketQueue     audioq;
  DECLARE_ALIGNED(16, uint8_t, audio_buf[(AVCODEC_MAX_AUDIO_FRAME_SIZE * 3) / 2]);
  unsigned int    audio_buf_size;
  unsigned int    audio_buf_index;
  AVPacket        audio_pkt;
  uint8_t         *audio_pkt_data;
  int             audio_pkt_size;
  int             audio_hw_buf_size;  
  double          audio_diff_cum; /* used for AV difference average computation */
  double          audio_diff_avg_coef;
  double          audio_diff_threshold;
  int             audio_diff_avg_count;
  double          audio_current_pts;
  int64_t         audio_current_pts_time;

  double          frame_timer;
  double          frame_last_pts;
  double          frame_last_delay;
  double          video_clock; ///<pts of last decoded frame / predicted pts of next decoded frame
  double          video_current_pts; ///<current displayed pts (different from video_clock if frame fifos are used)
  int64_t         video_current_pts_time;  ///<time (av_gettime) at which we updated video_current_pts - used to have running video pts
  AVStream        *video_st;
  PacketQueue     videoq;
  VideoPicture    pictq[VIDEO_PICTURE_QUEUE_SIZE];
  int             pictq_size, pictq_rindex, pictq_windex;

  SDL_mutex       *pictq_mutex;
  SDL_cond        *pictq_cond;
  SDL_Thread      *parse_tid;
  SDL_Thread      *video_tid;

  char            filename[1024];
  int             width;
  int             height;
  int             quit;
  int             pause_curr;
  int             pause_last;
  int             pause_back;
  int             step;

  struct SwsContext *img_convert_ctx;

} VideoState;

enum {
  AV_SYNC_AUDIO_MASTER,
  AV_SYNC_VIDEO_MASTER,
  AV_SYNC_EXTERNAL_MASTER,
};

/* options passed by the caller */
static intmax_t time_s = 0;
static intmax_t time_d = 0;
static int screen_w = 0;
static int screen_h = 0;
static int full = 0;
static int loop = 0;

static SDL_Surface *screen;
/* since we have only one decoding thread, we can use a global
   variable instead of a thread local variable */
static VideoState *global_video_state;

static AVPacket flush_pkt;

static int packet_queue_put(PacketQueue *q, AVPacket *pkt) {

  AVPacketList *pkt1;

  /* duplicate the packet */
  if(pkt != &flush_pkt && av_dup_packet(pkt) < 0) {
    return -1;
  }

  pkt1 = av_malloc(sizeof(AVPacketList));
  if (!pkt1)
    return -1;
  pkt1->pkt = *pkt;
  pkt1->next = NULL;
  

  SDL_LockMutex(q->mutex);

  if (!q->last_pkt)
    q->first_pkt = pkt1;
  else
    q->last_pkt->next = pkt1;
  q->last_pkt = pkt1;
  q->nb_packets++;
  q->size += pkt1->pkt.size;

  /* OO: should duplicate packet data in DV case */
  SDL_CondSignal(q->cond);
  SDL_UnlockMutex(q->mutex);
  return 0;
}

/* packet queue handling */
static void packet_queue_init(PacketQueue *q) {
  memset(q, 0, sizeof(PacketQueue));
  q->mutex = SDL_CreateMutex();
  q->cond = SDL_CreateCond();
  packet_queue_put(q, &flush_pkt);
}

static void packet_queue_flush(PacketQueue *q) {
  AVPacketList *pkt, *pkt1;
  
  SDL_LockMutex(q->mutex);
  for(pkt = q->first_pkt; pkt != NULL; pkt = pkt1) {
    pkt1 = pkt->next;
    av_free_packet(&pkt->pkt);
    av_freep(&pkt);
  }
  q->last_pkt = NULL;
  q->first_pkt = NULL;
  q->nb_packets = 0;
  q->size = 0;
  SDL_UnlockMutex(q->mutex);
}

static void packet_queue_end(PacketQueue *q)
{
    packet_queue_flush(q);
    SDL_DestroyMutex(q->mutex);
    SDL_DestroyCond(q->cond);
}

static void packet_queue_quit(PacketQueue *q)
{
    SDL_LockMutex(q->mutex);
    q->quit = 1;
    SDL_CondSignal(q->cond);
    SDL_UnlockMutex(q->mutex);
}

/* return < 0 if aborted, 0 if no packet and > 0 if packet.  */
static int packet_queue_get(PacketQueue *q, AVPacket *pkt, int block) {

  AVPacketList *pkt1;
  int ret;

  SDL_LockMutex(q->mutex);
  
  for(;;) {
    if(q->quit) {
      ret = -1;
      break;
    }

    pkt1 = q->first_pkt;
    if (pkt1) {
      q->first_pkt = pkt1->next;
      if (!q->first_pkt)
	q->last_pkt = NULL;
      q->nb_packets--;
      q->size -= pkt1->pkt.size;
      *pkt = pkt1->pkt;
      av_free(pkt1);
      ret = 1;
      break;
    } else if (!block) {
      ret = 0;
      break;
    } else {
      SDL_CondWait(q->cond, q->mutex);
    }
  }
  SDL_UnlockMutex(q->mutex);
  return ret;
}



static double get_audio_clock(VideoState *is) {
    if (is->pause_curr) {
        return is->audio_current_pts;
    } else {
        return is->audio_current_pts_time + av_gettime() / 1000000.0;
    }

}

static double get_video_clock(VideoState *is) {
    if (is->pause_curr) {
        return is->video_current_pts;
    } else {
       return is->video_current_pts_time + av_gettime() / 1000000.0;
    }
}

static double get_external_clock(VideoState *is) {

    int64_t ti;
    ti = av_gettime();
    return is->external_clock + ((ti - is->external_clock_time) * 1e-6);
}

static double get_master_clock(VideoState *is) {

    double val;

    if (is->av_sync_type == AV_SYNC_VIDEO_MASTER) {
        if (is->video_st)
            val = get_video_clock(is);
        else
            val = get_audio_clock(is);
    } else if (is->av_sync_type == AV_SYNC_AUDIO_MASTER) {
        if (is->audio_st)
            val = get_audio_clock(is);
        else
            val = get_video_clock(is);
    } else {
        val = get_external_clock(is);
    }
    return val;

}

/* return the new audio buffer size (samples can be added or deleted
   to get better sync if video or external master clock) */
static int synchronize_audio(VideoState *is, short *samples,
		             int samples_size, double pts) {
  int n;
  double ref_clock;
  
  n = 2 * is->audio_st->codec->channels;
  
  /* if not master, then we try to remove or add samples to correct the clock */
  if(((is->av_sync_type == AV_SYNC_VIDEO_MASTER && is->video_st) ||
      is->av_sync_type == AV_SYNC_EXTERNAL_MASTER)) {
    double diff, avg_diff;
    int wanted_size, min_size, max_size, nb_samples;
    
    ref_clock = get_master_clock(is);
    diff = get_audio_clock(is) - ref_clock;

    if(diff < AV_NOSYNC_THRESHOLD) {
      // accumulate the diffs
      is->audio_diff_cum = diff + is->audio_diff_avg_coef * is->audio_diff_cum;
      if(is->audio_diff_avg_count < AUDIO_DIFF_AVG_NB) {
        /* not enough measures to have a correct estimate */
	is->audio_diff_avg_count++;
      } else {
        /* estimate the A-V difference */
	avg_diff = is->audio_diff_cum * (1.0 - is->audio_diff_avg_coef);

	if(fabs(avg_diff) >= is->audio_diff_threshold) {
	  wanted_size = samples_size + ((int)(diff * is->audio_st->codec->sample_rate) * n);
          nb_samples = samples_size / n;

          min_size = ((nb_samples * (100 - SAMPLE_CORRECTION_PERCENT_MAX)) / 100) * n;
          max_size = ((nb_samples * (100 + SAMPLE_CORRECTION_PERCENT_MAX)) / 100) * n;
	  if(wanted_size < min_size) {
	    wanted_size = min_size;
	  } else if (wanted_size > max_size) {
	    wanted_size = max_size;
	  }

          /* add or remove samples to correction the synchro */
	  if(wanted_size < samples_size) {
	    /* remove samples */
	    samples_size = wanted_size;
	  } else if(wanted_size > samples_size) {
	    uint8_t *samples_end, *q;
	    int nb;

	    /* add samples by copying final sample*/
	    nb = (samples_size - wanted_size);
	    samples_end = (uint8_t *)samples + samples_size - n;
	    q = samples_end + n;
	    while(nb > 0) {
	      memcpy(q, samples_end, n);
	      q += n;
	      nb -= n;
	    }
	    samples_size = wanted_size;
	  }
	}
#if 0
                printf("diff=%f adiff=%f sample_diff=%d apts=%0.3f vpts=%0.3f %f\n",
                       diff, avg_diff, samples_size - samples_size1,
                       is->audio_clock, is->video_clock, is->audio_diff_threshold);
#endif
      }
    } else {
      /* difference is TOO big; reset A-V filter */
      is->audio_diff_avg_count = 0;
      is->audio_diff_cum = 0;
    }
  }
  return samples_size;
}

/* decode one audio frame and returns its uncompressed size */
static int audio_decode_frame(VideoState *is, uint8_t *audio_buf, int buf_size, double *pts_ptr) {

  AVPacket *pkt = &is->audio_pkt;
  AVCodecContext *dec = is->audio_st->codec;
  int len1, data_size, n;
  double pts;

  for(;;) {
    /* NOTE: the audio packet can contain several frames */
    while(is->audio_pkt_size > 0) {
      data_size = buf_size;

      len1 = avcodec_decode_audio3(is->audio_st->codec, 
				   (int16_t *)audio_buf, &data_size, 
				   &is->audio_pkt);
      if(len1 < 0) {
	/* if error, skip frame */
	is->audio_pkt_size = 0;
	break;
      }
      is->audio_pkt_data += len1;
      is->audio_pkt_size -= len1;
      if(data_size <= 0) {
	/* No data yet, get more frames */
	continue;
      }
      pts = is->audio_clock;
      *pts_ptr = pts;
      n = 2 * dec->channels;
      is->audio_clock += (double)data_size /
	(double)(n * dec->sample_rate);

      /* We have data, return it and come back for more later */
      return data_size;
    }

    /* free the current packet */
    if(pkt->data)
      av_free_packet(pkt);

    if(is->pause_curr || is->quit) {
      return -1;
    }

    /* next packet */
    if(packet_queue_get(&is->audioq, pkt, 1) < 0) {
      return -1;
    }
    if(pkt->data == flush_pkt.data) {
      avcodec_flush_buffers(is->audio_st->codec);
      continue;
    }

    is->audio_pkt_data = pkt->data;
    is->audio_pkt_size = pkt->size;

    /* if update, update the audio clock w/pts */
    if(pkt->pts != AV_NOPTS_VALUE) {
      is->audio_clock = av_q2d(is->audio_st->time_base)*pkt->pts;
    }
  }
}

/* prepare a new audio buffer */
static void audio_callback(void *userdata, Uint8 *stream, int len) {
  VideoState *is = (VideoState *)userdata;
  int len1, audio_size;
  double pts;

  while(len > 0) {
    if(is->audio_buf_index >= is->audio_buf_size) {
      /* We have already sent all our data; get more */
      audio_size = audio_decode_frame(is, is->audio_buf, sizeof(is->audio_buf), &pts);
      if(audio_size < 0) {
	/* If error, output silence */
	is->audio_buf_size = 1024;
	memset(is->audio_buf, 0, is->audio_buf_size);
      } else {
	audio_size = synchronize_audio(is, (int16_t *)is->audio_buf,
				       audio_size, pts);
	is->audio_buf_size = audio_size;
      }
      is->audio_buf_index = 0;
    }
    len1 = is->audio_buf_size - is->audio_buf_index;
    if(len1 > len)
      len1 = len;
    memcpy(stream, (uint8_t *)is->audio_buf + is->audio_buf_index, len1);
    len -= len1;
    stream += len1;
    is->audio_buf_index += len1;
  }
}

static Uint32 sdl_refresh(Uint32 interval, void *opaque) {
  VideoState *is= opaque;
  SDL_Event event;
  event.type = FF_REFRESH_EVENT;
  event.user.data1 = opaque;
  SDL_PushEvent(&event);

  return 0; /* 0 means stop timer */
}

/* schedule a video refresh in 'delay' ms */
static void schedule_refresh(VideoState *is, int delay) {
  SDL_AddTimer(delay, sdl_refresh, is);
}

static void video_display(VideoState *is) {
  SDL_Rect rect;
  VideoPicture *vp;
  AVPicture pict;
  float aspect_ratio;
  int w, h, x, y;
  int i;

  vp = &is->pictq[is->pictq_rindex];
  if(vp->bmp) {
    if(is->video_st->codec->sample_aspect_ratio.num == 0) {
      aspect_ratio = 0;
    } else {
      aspect_ratio = av_q2d(is->video_st->codec->sample_aspect_ratio) *
	is->video_st->codec->width / is->video_st->codec->height;
    }
    if(aspect_ratio <= 0.0) {
      aspect_ratio = (float)is->video_st->codec->width /
	(float)is->video_st->codec->height;
    }
    // apparently this assumption is bad
    h = screen->h;
    w = ((int)rint(h * aspect_ratio)) & -3;
    if(w > screen->w) {
      w = screen->w;
      h = ((int)rint(w / aspect_ratio)) & -3;
    }
    x = (screen->w - w) / 2;
    y = (screen->h - h) / 2;
    rect.x = x;
    rect.y = y;
    rect.w = w;
    rect.h = h;
    SDL_DisplayYUVOverlay(vp->bmp, &rect);
  }
}

static void video_refresh_timer(void *userdata) {

  VideoState *is = (VideoState *)userdata;
  VideoPicture *vp;

  double actual_delay, delay, sync_threshold, ref_clock, diff;
  
  if(is->video_st) {
    if(is->pictq_size == 0) {
      schedule_refresh(is, 1);
    } else {
      vp = &is->pictq[is->pictq_rindex];

      is->video_current_pts = vp->pts;
      //is->video_current_pts_time = av_gettime();
      is->video_current_pts_time = is->video_current_pts - (av_gettime()/1000000.0);

      delay = vp->pts - is->frame_last_pts; /* the pts from last time */
      if(delay <= 0 || delay >= 1.0) {
	/* if incorrect delay, use previous one */
	delay = is->frame_last_delay;
      }
      /* save for next time */
      is->frame_last_delay = delay;
      is->frame_last_pts = vp->pts;

      /* update delay to sync to audio if not master source */
      if(is->av_sync_type != AV_SYNC_VIDEO_MASTER) {
	ref_clock = get_master_clock(is);
	diff = vp->pts - ref_clock;
	
	/* Skip or repeat the frame. Take delay into account
	   FFPlay still doesn't "know if this is the best guess." */
	sync_threshold = (delay > AV_SYNC_THRESHOLD) ? delay : AV_SYNC_THRESHOLD;
	if(fabs(diff) < AV_NOSYNC_THRESHOLD) {
	  if(diff <= -sync_threshold) {
	    delay = 0;
	  } else if(diff >= sync_threshold) {
	    delay = 2 * delay;
	  }
	}
      }

      is->frame_timer += delay;
      /* computer the REAL delay */
      actual_delay = is->frame_timer - (av_gettime() / 1000000.0);
      if(actual_delay < 0.010) {
	/* Really it should skip the picture instead */
	actual_delay = 0.010;
      }
      schedule_refresh(is, (int)(actual_delay * 1000 + 0.5));

      /* show the picture! */
      video_display(is);
      
      /* update queue for next picture! */
      if(++is->pictq_rindex == VIDEO_PICTURE_QUEUE_SIZE) {
	is->pictq_rindex = 0;
      }
      SDL_LockMutex(is->pictq_mutex);
      is->pictq_size--;
      SDL_CondSignal(is->pictq_cond);
      SDL_UnlockMutex(is->pictq_mutex);
    }
  } else {
    schedule_refresh(is, 100);
  }
}
      
static void alloc_picture(void *userdata) {

  VideoState *is = (VideoState *)userdata;
  VideoPicture *vp;

  vp = &is->pictq[is->pictq_windex];
  if(vp->bmp) {
    // we already have one make another, bigger/smaller
    SDL_FreeYUVOverlay(vp->bmp);
  }
  // Allocate a place to put our YUV image on that screen
  vp->bmp = SDL_CreateYUVOverlay(is->video_st->codec->width,
				 is->video_st->codec->height,
				 SDL_YV12_OVERLAY,
				 screen);
  vp->width = is->video_st->codec->width;
  vp->height = is->video_st->codec->height;
  
  SDL_LockMutex(is->pictq_mutex);
  vp->allocated = 1;
  SDL_CondSignal(is->pictq_cond);
  SDL_UnlockMutex(is->pictq_mutex);

}

static int queue_picture(VideoState *is, AVFrame *src_frame, double pts) {

  VideoPicture *vp;

  /* wait until we have space for a new pic */
  SDL_LockMutex(is->pictq_mutex);
  while(is->pictq_size >= VIDEO_PICTURE_QUEUE_SIZE &&
	!is->quit) {
    SDL_CondWait(is->pictq_cond, is->pictq_mutex);
  }
  SDL_UnlockMutex(is->pictq_mutex);

  if(is->quit)
    return -1;

  // windex is set to 0 initially
  vp = &is->pictq[is->pictq_windex];

  /* allocate or resize hardware picture buffer */
  if(!vp->bmp ||
    vp->width != is->video_st->codec->width ||
    vp->height != is->video_st->codec->height) {

    SDL_Event event;

    vp->allocated = 0;

    /* the allocation must be done in the main thread to avoid locking problems */
    event.type = FF_ALLOC_EVENT;
    event.user.data1 = is;
    SDL_PushEvent(&event);

    /* wait until we have a picture allocated */
    SDL_LockMutex(is->pictq_mutex);
    while(!vp->allocated && !is->quit) {
      SDL_CondWait(is->pictq_cond, is->pictq_mutex);
    }
    /* if the queue is aborted, we have to pop the pending ALLOC event or wait for the allocation to complete */
    if (is->quit && SDL_PeepEvents(&event, 1, SDL_GETEVENT, SDL_EVENTMASK(FF_ALLOC_EVENT)) != 1) {
        while (!vp->allocated) {
            SDL_CondWait(is->pictq_cond, is->pictq_mutex);
        }
    }

    SDL_UnlockMutex(is->pictq_mutex);

    if(is->quit) {
      return -1;
    }
  }
  /* We have a place to put our picture on the queue */
  /* If we are skipping a frame, do we set this to null 
     but still return vp->allocated = 1? */


  if(vp->bmp) {
    AVPicture pict;

    /* get a pointer on the bitmap */
    SDL_LockYUVOverlay(vp->bmp);
    
    /* point pict at the queue */
    memset(&pict,0,sizeof(AVPicture));
    pict.data[0] = vp->bmp->pixels[0];
    pict.data[1] = vp->bmp->pixels[2];
    pict.data[2] = vp->bmp->pixels[1];
    
    pict.linesize[0] = vp->bmp->pitches[0];
    pict.linesize[1] = vp->bmp->pitches[2];
    pict.linesize[2] = vp->bmp->pitches[1];

    is->img_convert_ctx = sws_getCachedContext(is->img_convert_ctx,
        vp->width, vp->height, vp->pix_fmt, vp->width, vp->height,
        PIX_FMT_YUV420P, sws_flags, NULL, NULL, NULL);
    if(is->img_convert_ctx == NULL) {
	fprintf(stderr, "Cannot initialize the conversion context!\n");
	return 1;
    }

    sws_scale(is->img_convert_ctx, (const uint8_t * const*)src_frame->data, src_frame->linesize,
	      0, vp->height, pict.data, pict.linesize);

    /* update the bitmap content */
    SDL_UnlockYUVOverlay(vp->bmp);

    vp->pts = pts;

    /* now we inform our display thread that we have a pic ready */
    if(++is->pictq_windex == VIDEO_PICTURE_QUEUE_SIZE) {
      is->pictq_windex = 0;
    }
    SDL_LockMutex(is->pictq_mutex);
    is->pictq_size++;
    SDL_UnlockMutex(is->pictq_mutex);
  }
  return 0;
}

static double synchronize_video(VideoState *is, AVFrame *src_frame, double pts) {

  double frame_delay;

  if(pts != 0) {
    /* if we have pts, set video clock to it */
    is->video_clock = pts;
  } else {
    /* if we aren't given a pts, set it to the clock */
    pts = is->video_clock;
  }
  /* update the video clock */
  frame_delay = av_q2d(is->video_st->codec->time_base);
  /* if we are repeating a frame, adjust clock accordingly */
  frame_delay += src_frame->repeat_pict * (frame_delay * 0.5);
  is->video_clock += frame_delay;
  return pts;
}

static int get_video_frame(VideoState *is, AVFrame *frame, int64_t *pts, AVPacket *pkt)
{
    int got_picture, i;

    if (packet_queue_get(&is->videoq, pkt, 1) < 0)
        return -1;

    if (pkt->data == flush_pkt.data) {
        avcodec_flush_buffers(is->video_st->codec);

        SDL_LockMutex(is->pictq_mutex);
        while (is->pictq_size && !is->videoq.quit) {
            SDL_CondWait(is->pictq_cond, is->pictq_mutex);
        }
        SDL_UnlockMutex(is->pictq_mutex);

        is->frame_last_pts = AV_NOPTS_VALUE;
        is->frame_last_delay = 0;
        is->frame_timer = (double)av_gettime() / 1000000.0;
        return 0;
    }

    avcodec_decode_video2(is->video_st->codec, frame, &got_picture, pkt);

    if (got_picture) {

        *pts = frame->best_effort_timestamp;
        if (!*pts) {
            *pts = frame->pkt_dts;
        }

        if (*pts == AV_NOPTS_VALUE) {
            *pts = 0;
        }

    }
    return 0;
}

static void do_pause(VideoState *is) {
    if (is->pause_curr) {
        is->frame_timer += av_gettime() / 1000000.0 + is->video_current_pts_time - is->video_current_pts;
        if(is->pause_back != AVERROR(ENOSYS)){
            is->video_current_pts = is->video_current_pts_time + av_gettime() / 1000000.0;
        }
        is->video_current_pts_time = is->video_current_pts - av_gettime() / 1000000.0;
    }

    is->pause_curr = !is->pause_curr;
    is->step = 0;
}


/* These are called whenever we allocate a frame
 * buffer. We use this to store the global_pts in
 * a frame at the time it is allocated.
 */
static int in_get_buffer(struct AVCodecContext *c, AVFrame *pic) {
  int ret = avcodec_default_get_buffer(c, pic);
  uint64_t *pts = av_malloc(sizeof(uint64_t));
  *pts = AV_NOPTS_VALUE;
  pic->opaque = pts;
  return ret;
}

static void in_release_buffer(struct AVCodecContext *c, AVFrame *pic) {
  if(pic) av_freep(&pic->opaque);
  avcodec_default_release_buffer(c, pic);
}

static int video_thread(void *arg) {
  VideoState *is = (VideoState *)arg;
  AVFrame *p_frame= avcodec_alloc_frame();
  int64_t pts_int = AV_NOPTS_VALUE, pos = -1;
  AVPacket pkt, *packet = &pkt;
  double pts;
  int ret;

  for(;;) {

    while (is->pause_curr && !is->videoq.quit)
       SDL_Delay(10);

       ret = get_video_frame(is, p_frame, &pts_int, &pkt);

       pos = pkt.pos;
       av_free_packet(&pkt);

       if (ret < 0) 
          goto the_end;

       pts = pts_int*av_q2d(is->video_st->time_base);
       ret = queue_picture(is, p_frame, pts);

       if (ret < 0)
          goto the_end;

       if (is->step)
          do_pause(is);
  }
the_end:
  av_free(p_frame);
  return 0;
}


/* open a given stream. Return 0 if OK */
static int stream_open(VideoState *is, int stream_index) {

  AVFormatContext *avf_ctx = is->avf_ctx;
  AVCodecContext *avc_ctx;
  AVCodec *codec;
  SDL_AudioSpec wanted_spec, spec;

  if(stream_index < 0 || stream_index >= avf_ctx->nb_streams) {
    return -1;
  }

  // Get a pointer to the codec context for the video stream
  avc_ctx = avf_ctx->streams[stream_index]->codec;

    /* prepare audio output */
    if (avc_ctx->codec_type == AVMEDIA_TYPE_AUDIO) {
        if (avc_ctx->channels > 0) {
            avc_ctx->request_channels = FFMIN(2, avc_ctx->channels);
        } else {
            avc_ctx->request_channels = 2;
        }
    }

  codec = avcodec_find_decoder(avc_ctx->codec_id);
  if(!codec || (avcodec_open(avc_ctx, codec) < 0)) {
      fprintf(stderr, "Unsupported codec!\n");
      return -1;
  }

  /* prepare audio output */
  if(avc_ctx->codec_type == AVMEDIA_TYPE_AUDIO) {
    if(avc_ctx->sample_rate <= 0 || avc_ctx->channels <= 0){
        fprintf(stderr, "Invalid sample rate or channel count\n");
        return -1;
    }

    // Set audio settings from codec info
    wanted_spec.freq = avc_ctx->sample_rate;
    wanted_spec.format = AUDIO_S16SYS;
    wanted_spec.channels = avc_ctx->channels;
    wanted_spec.silence = 0;
    wanted_spec.samples = SDL_AUDIO_BUFFER_SIZE;
    wanted_spec.callback = audio_callback;
    wanted_spec.userdata = is;
    
    if(SDL_OpenAudio(&wanted_spec, &spec) < 0) {
        fprintf(stderr, "SDL_OpenAudio: %s\n", SDL_GetError());
        return -1;
    }
    is->audio_hw_buf_size = spec.size;
    //is->audio_src_fmt= AV_SAMPLE_FMT_S16;
  }


  avf_ctx->streams[stream_index]->discard = AVDISCARD_DEFAULT;
  switch(avc_ctx->codec_type) {
  case AVMEDIA_TYPE_AUDIO:
    is->audio_stream = stream_index;
    is->audio_st = avf_ctx->streams[stream_index];
    is->audio_buf_size = 0;
    is->audio_buf_index = 0;
    
    /* averaging filter for audio sync */
    is->audio_diff_avg_coef = exp(log(0.01 / AUDIO_DIFF_AVG_NB));
    is->audio_diff_avg_count = 0;
    /* Correct audio only if larger error than this */
    is->audio_diff_threshold = 2.0 * SDL_AUDIO_BUFFER_SIZE / avc_ctx->sample_rate;

    memset(&is->audio_pkt, 0, sizeof(is->audio_pkt));
    packet_queue_init(&is->audioq);
    SDL_PauseAudio(0);
    break;
  case CODEC_TYPE_VIDEO:
    is->video_stream = stream_index;
    is->video_st = avf_ctx->streams[stream_index];

    is->frame_timer = (double)av_gettime() / 1000000.0;
    is->frame_last_delay = 40e-3;
    is->video_current_pts_time = av_gettime();

    packet_queue_init(&is->videoq);
    is->video_tid = SDL_CreateThread(video_thread, is);
    avc_ctx->get_buffer = in_get_buffer;
    avc_ctx->release_buffer = in_release_buffer;

    break;
  default:
    break;
  }
  return 0;
}

/* safety */
static void stream_component_close(VideoState *is, int stream_index)
{
    AVFormatContext *avf_ctx = is->avf_ctx;
    AVCodecContext *avctx;

    if (stream_index < 0 || stream_index >= avf_ctx->nb_streams)
        return;
    avctx = avf_ctx->streams[stream_index]->codec;

    switch(avctx->codec_type) {
    case AVMEDIA_TYPE_VIDEO:
        packet_queue_quit(&is->videoq);
        /* note: we also signal this mutex to make sure we deblock the video thread in all cases */
        SDL_LockMutex(is->pictq_mutex);
        SDL_CondSignal(is->pictq_cond);
        SDL_UnlockMutex(is->pictq_mutex);
        SDL_WaitThread(is->video_tid, NULL);
        packet_queue_end(&is->videoq);
        break;
    case AVMEDIA_TYPE_AUDIO:
        packet_queue_quit(&is->audioq);
        SDL_CloseAudio();
        packet_queue_end(&is->audioq);
        break;
    default:
        break;
    }

    avf_ctx->streams[stream_index]->discard = AVDISCARD_ALL;
    avcodec_close(avctx);
    switch(avctx->codec_type) {
    case AVMEDIA_TYPE_VIDEO:
        is->video_st = NULL;
        is->video_stream = -1;
        break;
    case AVMEDIA_TYPE_AUDIO:
        is->audio_st = NULL;
        is->audio_stream = -1;
        break;
    default:
        break;
    }

}

void stream_seek(VideoState *is, int64_t pos, int64_t rel) {

  if(!is->seek_req) {
    is->seek_pos = pos;
    is->seek_rel = rel;
    is->seek_flags &= ~AVSEEK_FLAG_BYTE;
    is->seek_req = 1;
  }
}

static int decode_interrupt_cb(void) {
  return (global_video_state && global_video_state->quit);
}

/* this thread gets the stream from the disk */
static int decode_thread(void *arg) {

  VideoState *is = (VideoState *)arg;
  AVFormatContext *avf_ctx = NULL;
  AVPacket pkt1, *packet = &pkt1;

  int video_index = -1;
  int audio_index = -1;
  int pkt_user_range = 0;
  int i;
  int eof = 0;

  is->video_stream=-1;
  is->audio_stream=-1;

  global_video_state = is;
  // This will interrupt blocking functions when we quit!
  avio_set_interrupt_cb(decode_interrupt_cb);

  // Open video file
  if(avformat_open_input(&avf_ctx, is->filename, NULL, NULL)!=0)
     goto fail; // Couldn't open file

  is->avf_ctx = avf_ctx;
  
  // Retrieve stream information
  if(av_find_stream_info(avf_ctx)<0)
     goto fail; // Couldn't find stream information
  
  // Dump information about file onto standard error
  av_dump_format(avf_ctx, 0, is->filename, 0);

  // If seeking requested, we execute it
  if (time_s != 0) {

     int64_t timestep;
     timestep = time_s;

     // Add the stream starttime
     timestep += time_s;

     if (avformat_seek_file(avf_ctx, -1, INT64_MIN, timestep, INT64_MAX, 0) < 0) {
         fprintf(stderr, "%s: could not seek to position %0.3f\n",
         is->filename, (double)timestep / AV_TIME_BASE);
     }
  }

  for(i=0; i<avf_ctx->nb_streams; i++) {
  // Find the first video stream
    if(avf_ctx->streams[i]->codec->codec_type==CODEC_TYPE_VIDEO &&
       video_index < 0) {
      video_index=i;
    }
  // Find the first audio stream
    if(avf_ctx->streams[i]->codec->codec_type==CODEC_TYPE_AUDIO &&
       audio_index < 0) {
      audio_index=i;
    }
  }
  if(audio_index >= 0) {
    stream_open(is, audio_index);
  }
  if(video_index >= 0) {
    stream_open(is, video_index);
  }   

  if(is->video_stream < 0 || is->audio_stream < 0) {
    fprintf(stderr, "%s: could not open codecs\n", is->filename);
    goto fail;
  }

  // the main decoding loop
  for(;;) {
    if(is->quit) {
      break;
    }
    if (is->pause_curr != is->pause_last) {
      is->pause_last = is->pause_curr;
      if(is->pause_curr) {
        is->pause_back = av_read_pause(avf_ctx);
      }
      else {
         av_read_play(avf_ctx);
      }
    } 

    // the seeking subroutine
    if(is->seek_req) {
      int stream_index= -1;
      int64_t seek_target = is->seek_pos;

      if     (is->video_stream >= 0) stream_index = is->video_stream;
      else if(is->audio_stream >= 0) stream_index = is->audio_stream;


      if(stream_index>=0){
	seek_target= av_rescale_q(seek_target, AV_TIME_BASE_Q, avf_ctx->streams[stream_index]->time_base);
      }
      if(av_seek_frame(is->avf_ctx, stream_index, seek_target, is->seek_flags) < 0) {
	
	fprintf(stderr, "%s: error while seeking. target: %d, stream_index: %d\n", is->avf_ctx->filename, seek_target, stream_index);
      } else {
	if(is->audio_stream >= 0) {
	  packet_queue_flush(&is->audioq);
	  packet_queue_put(&is->audioq, &flush_pkt);
	}
	if(is->video_stream >= 0) {
	  packet_queue_flush(&is->videoq);
	  packet_queue_put(&is->videoq, &flush_pkt);
	}
      }
      is->seek_req = 0;
    }

    /* if the queue are full, no need to read more */
    if((is->audioq.size > MAX_AUDIOQ_SIZE || is->audio_stream<0) ||
       (is->videoq.size > MAX_VIDEOQ_SIZE || is->video_stream<0)) {
      /* wait 10 ms */
      SDL_Delay(10);
      continue;
    }

    /* check if we are at the end */
    if(eof) {
            if(is->video_stream >= 0){
                av_init_packet(packet);
                packet->data=NULL;
                packet->size=0;
                packet->stream_index= is->video_stream;
                packet_queue_put(&is->videoq, packet);
            }
            SDL_Delay(10);
            if(is->audioq.size + is->videoq.size == 0){
                if(loop == 1 ){
                    stream_seek(is, time_s != 0 ? time_s : 0, 0);
                }else if(loop == 0 ){
                    goto fail;
                }
            }
            eof=0;
            continue;
    }


    if(av_read_frame(is->avf_ctx, packet) < 0) {

        if(url_feof(avf_ctx->pb) || 
           av_read_frame(is->avf_ctx, packet) == AVERROR_EOF) {
             eof=1;
        }    
        if(avf_ctx->pb && avf_ctx->pb->error) {
             break;
        }
        SDL_Delay(100); /* wait for user event */
        continue;
    }

    /* check if packet is in play range specified by user, then queue, otherwise discard */
    if (time_d != 0) {
    pkt_user_range = 
       (packet->pts - avf_ctx->streams[packet->stream_index]->start_time) *
       av_q2d(avf_ctx->streams[packet->stream_index]->time_base) -
       (double)(time_s != 0 ? time_s : 0)/1000000
       <= ((double)time_d/1000000);
    }
 
    if(packet->stream_index == is->video_stream && pkt_user_range) {
      packet_queue_put(&is->videoq, packet);
    } else if(packet->stream_index == is->audio_stream && pkt_user_range) {
      packet_queue_put(&is->audioq, packet);
    } else {
      av_free_packet(packet);
    }
  }

  /* wait until the end */
  while(!is->quit) {
    SDL_Delay(100);
  }

 fail:
  {
    /* disable interrupting */
    global_video_state = NULL;
      
    if (is->video_stream >= 0)
        stream_component_close(is, is->video_stream);
    if (is->audio_stream >= 0)
        stream_component_close(is, is->audio_stream);
    if (is->avf_ctx) {
        av_close_input_file(is->avf_ctx);
        is->avf_ctx = NULL;  // safety
    } 

    avio_set_interrupt_cb(NULL);

    SDL_Event event;

    event.type = FF_QUIT_EVENT;
    event.user.data1 = is;
    SDL_PushEvent(&event);
  }
  return 0;
}


static void stream_close(VideoState *is) {

    VideoPicture *vp;
    int i;
    is->quit = 1;

    SDL_WaitThread(is->parse_tid, NULL);

    for(i=0;i<VIDEO_PICTURE_QUEUE_SIZE; i++) {
      vp = &is->pictq[i];

      if (vp->bmp) {
        SDL_FreeYUVOverlay(vp->bmp);
            vp->bmp = NULL;
      }
    }

    SDL_DestroyMutex(is->pictq_mutex);
    SDL_DestroyCond(is->pictq_cond);

    av_free(is);

}

static void do_exit(VideoState *is) {
    if (is) {
        stream_close(is);
    }
    av_lockmgr_register(NULL);

    SDL_Quit();
    av_log(NULL, AV_LOG_QUIET, "%s", "");
    return;
}

static void step_to_next_frame(VideoState *is)
{
    /* if the stream is paused unpause it, then step */
    if (is->pause_curr)
       do_pause(is);
    is->step = 1;
}


/* loop to handle an event sent by the GUI */
static int event_loop(VideoState *video_state) {
    SDL_Event event;
    double incr, pos;
    VideoState *is = video_state;

    if(full == 1) SDL_WM_ToggleFullScreen(screen);

    for(;;) {
    SDL_WaitEvent(&event);
    switch(event.type) {
    case SDL_KEYDOWN:
      switch(event.key.keysym.sym) {
         case SDLK_ESCAPE:
         case SDLK_q: // q: quit SDL loop
           do_exit(video_state);
           return 1;
           break;
         case SDLK_SPACE:
         case SDLK_p: //p: pause playing
           do_pause(video_state);
           break;
         case SDLK_s: //s: step to next frame
           step_to_next_frame(video_state);
           break;
         case SDLK_f: //f: fullscreen size
           SDL_WM_ToggleFullScreen(screen);
           break;
         case SDLK_o: // o: original picture size
           screen_w = is->video_st->codec->width;
           screen_h = is->video_st->codec->height;
           screen = SDL_SetVideoMode (screen_w, screen_h, 0, SDL_HWSURFACE|SDL_ASYNCBLIT|SDL_HWACCEL);
           break;
         case SDLK_h: // h: higher picture size
           screen_w = screen_w * 1.1;
           screen_h = screen_h * 1.1;
           screen = SDL_SetVideoMode (screen_w, screen_h, 0, SDL_HWSURFACE|SDL_ASYNCBLIT|SDL_HWACCEL);
           break;
         case SDLK_l: // l: lower picture size
           screen_w = screen_w / 1.1;
           screen_h = screen_h / 1.1;
           screen = SDL_SetVideoMode (screen_w, screen_h, 0, SDL_HWSURFACE|SDL_ASYNCBLIT|SDL_HWACCEL);
           break;
         case SDLK_LEFT:
	   incr = -10.0;
	   goto do_seek;
         case SDLK_RIGHT:
	   incr = 10.0;
	   goto do_seek;
         case SDLK_UP:
	   incr = 40.0;
	   goto do_seek;
         case SDLK_DOWN:
	   incr = -40.0;
	   goto do_seek;
         do_seek:
	   if(video_state) {
	     pos = get_master_clock(video_state);
	     pos += incr;
             stream_seek(video_state, (int64_t)(pos * AV_TIME_BASE), (int64_t)(incr * AV_TIME_BASE));
	   }
	   break;
         default:
	   break;
      }
      break;
    case SDL_VIDEORESIZE:
      screen = SDL_SetVideoMode(event.resize.w, event.resize.h, 0,
                                SDL_HWSURFACE|SDL_ASYNCBLIT|SDL_RESIZABLE|SDL_HWACCEL);
      screen_w = video_state->width = event.resize.w;
      screen_h = video_state->height = event.resize.h;
      break;
    case SDL_QUIT:
    case FF_QUIT_EVENT:
      do_exit(video_state);
      return 1;
      break;
    case FF_ALLOC_EVENT:
      alloc_picture(event.user.data1);
      break;
    case FF_REFRESH_EVENT:
      video_refresh_timer(event.user.data1);
      break;
    default:
      break;
    }
  }

}

static int lockmgr(void **mtx, enum AVLockOp op)
{
   switch(op) {
      case AV_LOCK_CREATE:
          *mtx = SDL_CreateMutex();
          if(!*mtx)
              return 1;
          return 0;
      case AV_LOCK_OBTAIN:
          return !!SDL_LockMutex(*mtx);
      case AV_LOCK_RELEASE:
          return !!SDL_UnlockMutex(*mtx);
      case AV_LOCK_DESTROY:
          SDL_DestroyMutex(*mtx);
          return 0;
   }
   return 1;
}

/* parse file and call the GUI loop 
                  (filename,   starttime, duration,  width, height, loop, full) */
int loadPlayerFile(char* name, int64_t s, int64_t d, int w, int h, int l, int f){
  printf("start, duration, loop, full %i, %i, %i, %i\n", s, d, l, f);
  int sdl_flags;
  VideoState  *is;
  int sw, sh;

  if(s > 0) time_s = 1000000 * s;
  if(d > 0) time_d = 1000000 * d; 

  if(l == 0 || l == 1) loop = l;
  if(f == 0 || f == 1) full = f;

  if(w>64 && h>64) {
      sw = screen_w = w;
      sh = screen_h = h;
  }
  else if(screen_w>64 && screen_h>64) {
      sw = screen_w;
      sh = screen_h;
  }
  else {
      sw = screen_w = 640;
      sh = screen_h = 480;
  }

  is = av_mallocz(sizeof(VideoState));
  if(!name) {
      fprintf(stderr, "A file must be specified\n");
      return 1;
  }

  stpncpy(is->filename, name, sizeof(is->filename));
  if (!is) {
      fprintf(stderr, "Failed to initialize VideoState!\n");
      do_exit(NULL);
  }

    /* register all codecs, demux and protocols */
#ifdef HAVE_LIBAVCODEC
    avcodec_register_all();
#endif
    av_register_all();

    /* set video flags */
#if !defined(__MINGW32__) && !defined(__APPLE__)
    sdl_flags = SDL_INIT_VIDEO | SDL_INIT_AUDIO | SDL_INIT_TIMER;
#else
    sdl_flags = SDL_INIT_VIDEO | SDL_INIT_AUDIO | SDL_INIT_TIMER | SDL_INIT_EVENTTHREAD;
#endif

  if(SDL_Init(sdl_flags)) {
      fprintf(stderr, "Could not initialize SDL - %s\n", SDL_GetError());
      return 1;
  }

  screen = SDL_SetVideoMode(sw, sh, 0, SDL_HWSURFACE|SDL_ASYNCBLIT|SDL_HWACCEL);

  sw = sh = 0;

  SDL_EventState(SDL_ACTIVEEVENT, SDL_IGNORE);
  SDL_EventState(SDL_SYSWMEVENT, SDL_IGNORE);
  SDL_EventState(SDL_USEREVENT, SDL_IGNORE);

  if (av_lockmgr_register(lockmgr)) {
     fprintf(stderr, "Could not initialize lock manager!\n");
     do_exit(NULL);
  }

  av_init_packet(&flush_pkt);
  flush_pkt.data = "FLUSH";

  is->pictq_mutex = SDL_CreateMutex();
  is->pictq_cond = SDL_CreateCond();

  schedule_refresh(is, 20);

  is->av_sync_type = DEFAULT_AV_SYNC_TYPE;
  is->parse_tid = SDL_CreateThread(decode_thread, is);
  if(!is->parse_tid) {
      av_free(is);
     return -1;
  }

  event_loop(is);

  // never return
  return 0;
}
